#<profile4xi profile = "10" />
# -------------------------------------------------------------------------------------------------
#<program4xi program = "_exit_function4xi.py">
#<abstract4xi abstract = "providing function _exit_function4xi to handle process exiting"/>
#<refer4xi reference = "util.py"/>
#<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Public API
# -------------------------------------------------------------------------------------------------
#<interface4xi interface = " 
__all__ = ['_exit_function4xi'] 
#"
#/>


# -------------------------------------------------------------------------------------------------
# Standard Imports                          
# -------------------------------------------------------------------------------------------------
#<standardize4xi standard = "
import os
import sys
from sys import exit
import threading                
import _thread
import _winapi   

import itertools
import weakref
import atexit
import logging 
#"
#/>

# -------------------------------------------------------------------------------------------------
# Multiprocessing Imports                          
# -------------------------------------------------------------------------------------------------
#<process4xi process = "
import _multiprocessing
try:                             # Python ones
    from multiprocessing.util import _exit_function
except:
    print("_exit_function4xi.py *** failed in from multiprocessing.util import _exit_function ***") 
#"
#/>

# -------------------------------------------------------------------------------------------------
# DragonXi imports                              
# -------------------------------------------------------------------------------------------------
#<dragonize4xi dragonization = "
#from active_children4xi import active_children4xi
from current_process4xi import list_children_current_process4xi
from is_exiting4xi import set_exiting4xi
from is_exiting4xi import get_exiting4xi
from print4xi import print4xi
from time4xi import time4xi
from _cleanup4xi import _cleanup4xi
from _run_finalizers4xi import _run_finalizers4xi
#"
#/>



# ---------------------------------------------------------------------------------------------
#<function4xi function = "exit-function4xi" usage = "
#   util.py
#       atexit.register(_exit_function)
#   process4xi.py:
#       _bootstrap:
#            _exit_function4xi()
#"> 
#<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
# ---------------------------------------------------------------------------------------------
def _exit_function4xi():
    '''
    exits
    '''

    # -----------------------------------------------------------------------------------------    
    # (Step 1/8): OK
    # -----------------------------------------------------------------------------------------        
 
    class2print = "_exit_function4xi"
    method2print = "_exit_function4xi"
    
    message2print = "{(Step 1a/8) started" + "; exit-function4xi-started: " + str(_exit_function4xi)  + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print) 
    # { _exit_function4xi : _exit_function4xi ; Step 1/5 start time: 15:23:40 09/02/09 FLE Daylight Time}
    
    status2exiting = get_exiting4xi()
    
    message2print = "{(Step 1b/8a) got" + "; status2exiting-got-by-exit-function4xi to: " + str(status2exiting)  + "; time: " + time4xi() + "}"
    print4xi(class2print, method2print, message2print)  
    
    if not status2exiting:
        # -----------------------------------------------------------------------------------------
        # (Step 2/8): OK
        # -----------------------------------------------------------------------------------------    
        set_exiting4xi(True)        #dragonized; added here
        
        message2print = "{(Step 2/8) performed" + "; set-exiting4xi-performed-by-exit-function4xi: " + str(set_exiting4xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)             
            
        # { _exit_function4xi : _exit_function4xi ; Step 2/5a performed: set_exiting4xi; time: 15:23:40 09/02/09 FLE Daylight Time} 
        
        # process shutting down
        # running all atexit finalizers with priority >= 0
    
        # { _exit_function4xi : _exit_function4xi ; Step 2/5b inform: process shutting down; time: 15:23:40 09/02/09 FLE Daylight Time} 
        # { _exit_function4xi : _exit_function4xi ; Step 2/5c inform: running finalizers ; time: 15:23:40 09/02/09 FLE Daylight Time}
            
        # -----------------------------------------------------------------------------------------
        # (Step 3/8): OK partly; no active children in tests so far
        # -----------------------------------------------------------------------------------------    
        
        _run_finalizers4xi(0)
        
        
        message2print = "{(Step 3/8) performed" + "; run-finalizers4xi-performed-by-exit-function4xi: " + str(_run_finalizers4xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)             
           
           
        #{ _exit_function4xi : _exit_function4xi ; Step 3/5a performed: _run_finalizers4xi; time: 15:23:40 09/02/09 FLE Daylight Time} 
            
        # -----------------------------------------------------------------------------------------    
        # (Step 4/8): 
        # -----------------------------------------------------------------------------------------        
        ack2terminate = False
        nack2terminate = True
        
        try:
        
            for childprocess2xi in active_children4xi():
            
                message2print = "{(Step 4/8a) got" + "; childprocess2xi-got-by-exit-function4xi to: " + str(childprocess2xi)  + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)  
                        
                try:
                    if childprocess2xi._daemonic:
                        # calling terminate() for daemon 
                        
                        message2print = "{(Step 4/8b) calling" + "; childprocess2xi-popen-terminate-calling-by-exit-function4xi: " + str(childprocess2xi._popen.terminate) + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)       
                                
                        childprocess2xi._popen.terminate()
            
                        message2print = "{performed" + "; childprocess2xi-popen-terminate-performed-by-exit-function4xi: " + str(childprocess2xi._popen.terminate) + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        
                        ack2terminate = True
                        
                except:
                        ack2terminate = False
                        nack2terminate = True
                        
                        message2print = "{yielded" + "; nack2terminate-yielded-by-exit-function4xi: " + str(nack2terminate)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print) 
                        
                                     
        except:
            message2print = "{yielded" + "; nack2terminate-yielded-by-exit-function4xi: " + str(nack2terminate)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)     
                                            
        # -----------------------------------------------------------------------------------------    
        # (Step 5/8): 
        # -----------------------------------------------------------------------------------------        
        ack2join = True     #ideal case
        nack2join = False   #ideal case
        
        try:
        
                
            for childprocess2xi in active_children4xi():
            
                try:
                    message2print = "{(Step 5/8a) got" + "; childprocess2xi-got-by-exit-function4xi to: " + str(childprocess2xi)  + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)  
                   
                
                    # calling join() for process
                    
                        
                    message2print = "{(Step 5/8b) calling" + "; childprocess2xi-join-calling-by-exit-function4xi: " + str(childprocess2xi.join) + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)       
                            
                    childprocess2xi.join()
            
                    message2print = "{performed" + "; childprocess2xi-join-performed-by-exit-function4xi: " + str(childprocess2xi.join) + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)             
                    
                   
                    message2print = "{(Step 5/8a) got" + "; childprocess2xi-got-by-exit-function4xi to: " + str(childprocess2xi)  + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)
                                   
                    
                       
                except:
                    ack2join = False  
                    nack2join = True                  
     
            
                    message2print = "{yielded" + "; nack2join-yielded-by-exit-function4xi: " + str(nack2join)  + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)                     
                       
        except:
            nack2join = True 
            
            message2print = "{yielded" + "; nack2join-yielded-by-exit-function4xi: " + str(nack2join)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
        
        if ((ack2join is False) or (nack2join is True)): #something went wrong
    
                
            message2print = "{warned" + "; nack2join-warned-by-exit-function4xi: " + " ** " + str(nack2join)  + " ** " + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)         
    
                                    
        # -----------------------------------------------------------------------------------------
        # (Step 6/8): OK 
        # -----------------------------------------------------------------------------------------        
        
        # running the remaining atexit finalizers
        
        _run_finalizers4xi()
    
        
        message2print = "{(Step 6/8) performed" + "; run_finalizers4xi-performed-by-exit-function4xi: " + str(_run_finalizers4xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)             
             
        # { _exit_function4xi : _exit_function4xi ; Step 4/5 performed: _run_finalizers4xi; time: 15:23:40 09/02/09 FLE Daylight Time}
         
        # ---------------------------------------------------------------------------------------------
        # (Step 7/8): OK
        # ---------------------------------------------------------------------------------------------    
    
        
        set_exiting4xi(False)       #dragonized; added here
        
        
        message2print = "{(Step 7/8) performed" + "; set-exiting4xi-performed-by-exit-function4xi: " + str(set_exiting4xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)             
             
        # { _exit_function4xi : _exit_function4xi ; Step 5/5a performed: set_exiting4xi; time: 15:23:40 09/02/09 FLE Daylight Time}    
        # { _exit_function4xi : _exit_function4xi ; Step 5/5b returning time: 15:23:40 09/02/09 FLE Daylight Time}

    else:
        message2print = "{(Step 1c/8a) got" + "; status2exiting-got-by-exit-function4xi to: " + str(status2exiting)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        # 
         
    # ---------------------------------------------------------------------------------------------
    # (Step 8/8): OK
    # ---------------------------------------------------------------------------------------------    
    message2print = "{(Step 8/8) returning" + "; exit-function4xi-returning: " + time4xi() + "}" 
    print4xi(class2print, method2print, message2print)    
    #     
                       
#</function4xi>

# -------------------------------------------------------------------------------------------------
# Settings JATKA, read >> http://docs.python.org/3.3/library/atexit.html
# -------------------------------------------------------------------------------------------------
#<set4xi setting = "
global function2register4exit

try:
    function2register4exit = _exit_function4xi
    # print("function2register4exit: ", function2register4exit) # <-- ok i.e. compiler finds the function
except:
    # print("_exit_function4xi *** failed to set function2register4exit ***") <-- never comers here!
    pass

try:
    atexit.register(function2register4exit)  # <-- global name _exit_function4xi is not defined!?
except:
    # print("_exit_function4xi.py *** failed in atexit.register(function2register4exit) ***")
    pass   
#"
#"/>


#</program4xi>

# -------------------------------------------------------------------------------------------------
# Recorded History                             
# ------------------------------------------------------------------------------------------------- 
#<record4xi record = "
#   test_exit_function4xi.txt
#   sense-test-exit-function4xi.xml
#   reason-test-exit-function4xi.xml
#   ace-test-exit-function4xi.xml
#"
#>

# -------------------------------------------------------------------------------------------------
# Legal Notice                             
# ------------------------------------------------------------------------------------------------- 
#<protect4xi protection = "copyright"
#   owner = "Small & Smart"
#   rights = "All rights reserved"
#   state = "California"
#   country = "United States of America"
#   references = "Legal notice at www.dragonxi.com; www.yourdragonxi.com; www.yourdragonxi.org; www.dragon2xi.com/; www.dragon4xi.com/; www.dragon8xi.com/ " 
#   codes = "Codes established in California and applicable to corporations in California"
#   compliance = "The California Law"
#   court = "Supreme Court of California, 350 McAllister Street, San Francisco, CA 94102-4783"
#/>
 



#<end4xi end = "
# -------------------------------------------------------------------------------------------------
# end of _exit_function4xi.py                             
# ------------------------------------------------------------------------------------------------- 
#"
#/>
 